import aws_cdk
import constructs


class Construct(constructs.Construct):

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        error_topic=None,
        lambda_directory=None,
        **kwargs
    ):
        super().__init__(scope, id, **kwargs)
        self.error_topic = error_topic

    def create_sns_topic(self, display_name):
        return aws_cdk.aws_sns.Topic(
            self, display_name,
            display_name=display_name,
        )

    @staticmethod
    def create_cloudwatch_math_expression(expression=None, label=None, using_metrics=None):
        return aws_cdk.aws_cloudwatch.MathExpression(
            expression=expression,
            label=label,
            using_metrics=using_metrics,
            period=aws_cdk.Duration.minutes(5),
        )

    def cloudwatch_math_sum(self, label=None, m1=None, m2=None):
        return self.create_cloudwatch_math_expression(
            label=label,
            expression="m1 + m2",
            using_metrics={"m1": m1, "m2": m2},
        )

    @staticmethod
    def create_cloudwatch_widget(title=None, stacked=True, left=None):
        return aws_cdk.aws_cloudwatch.GraphWidget(
            title=title, width=8, stacked=stacked, left=left
        )

    def create_cloudwatch_alarm(self, id=None, metric=None, threshold=1):
        alarm = aws_cdk.aws_cloudwatch.Alarm(
            self,
            id=id,
            metric=metric,
            threshold=threshold,
            evaluation_periods=6,
            datapoints_to_alarm=1,
            treat_missing_data=aws_cdk.aws_cloudwatch.TreatMissingData.NOT_BREACHING,
        )
        if self.error_topic:
            alarm.add_alarm_action(
                aws_cdk.aws_cloudwatch_actions.SnsAction(self.error_topic)
            )

    def create_cloudwatch_widgets(self):
        raise NotImplementedError


import aws_cdk
import aws_cdk.aws_apigatewayv2_alpha
import constructs

from . import well_architected_construct


class Api(well_architected_construct.Construct):

    def __init__(
        self,scope: constructs.Construct, id: str,
        error_topic=None,
        api=None,
        api_gateway_service_role=None,
        **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )
        self.api = api
        self.api_id = self.get_api_id(api)
        self.api_gateway_service_role = self.create_api_gateway_service_role(api_gateway_service_role)
        self.create_api_gateway_4xx_alarm()
        self.create_api_gateway_5xx_alarm()
        self.create_api_gateway_latency_alarm()
        aws_cdk.CfnOutput(self, f'{id} URL', value=api.url)

    def get_api_id(self, api):
        try:
            return api.api_id
        except AttributeError:
            return api.rest_api_id

    def create_api_gateway_service_role(self, iam_role):
        if iam_role:
            return iam_role
        if iam_role == False:
            return
        if iam_role is None:
            return aws_cdk.aws_iam.Role(
                self, 'ApiGatewayServiceRole',
                assumed_by=aws_cdk.aws_iam.ServicePrincipal('apigateway.amazonaws.com'),
            )

    def add_api_gateway_metric(self, metric_name: str = None, label: str = None,
            period=aws_cdk.Duration.seconds(900), statistic: str = 'sum',
        ):
        return aws_cdk.aws_cloudwatch.Metric(
            metric_name=metric_name,
            namespace="AWS/ApiGateway",
            dimensions_map={"ApiId": self.api_id},
            unit=aws_cdk.aws_cloudwatch.Unit.COUNT,
            label=label,
            statistic=statistic,
            period=period,
        )

    def create_api_gateway_4xx_alarm(self):
        return self.create_cloudwatch_alarm(
            id="4XXErrorsGreaterThanOnePercent",
            metric=self.create_cloudwatch_math_expression(
                expression="m1 / m2 * 100",
                label="api_gateway_4XX_errors_percentage",
                using_metrics={
                    "m1": self.add_api_gateway_metric(
                        metric_name='4XXError',
                        label='4XX_errors',
                    ),
                    "m2": self.add_api_gateway_metric(
                        metric_name='Count',
                        label='number_of_requests',
                    ),
                },
            ),
        )

    def create_api_gateway_5xx_alarm(self):
        return self.create_cloudwatch_alarm(
            id="5XXErrorsGreaterThanZero",
            metric=self.add_api_gateway_metric(
                metric_name="5XXError",
                label="api_gateway_5XX_errors",
                statistic="p99",
            ),
            threshold=0,
        )

    def create_api_gateway_latency_alarm(self):
        return self.create_cloudwatch_alarm(
            id="P99LatencyGreaterThanOneSecond",
            metric=self.add_api_gateway_metric(
                metric_name="Latency",
                label="api_gateway_latency",
                statistic="p99",
            ),
            threshold=1000,
        )

    @staticmethod
    def percentile_statistics():
        return ("p50", "p90", "p99",)

    @staticmethod
    def metric_names():
        return (f'{code}XXError' for code in ('4', '5'))

    def create_api_gateway_errors_widget(self):
        return self.create_cloudwatch_widget(
            title="api_gateway_errors",
            left=[
                self.add_api_gateway_metric(
                    metric_name=metric_name,
                    label=f'api_gateway_{metric_name}s',
                ) for metric_name in self.metric_names()
            ]
        )

    def create_api_gateway_latency_widget(self):
        return self.create_cloudwatch_widget(
            title="api_gateway_latency",
            left=[
                self.add_api_gateway_metric(
                    metric_name="Latency",
                    label=f'api_gateway_latency_{statistic}',
                    statistic=statistic,
                ) for statistic in self.percentile_statistics()
            ]
        )

    def create_api_gateway_number_of_requests_widget(self):
        return self.create_cloudwatch_widget(
            title="Requests",
            stacked=False,
            left=[
                self.add_api_gateway_metric(
                    metric_name="Count",
                    label="# Requests",
                )
            ]
        )

    def create_cloudwatch_widgets(self):
        return (
            self.create_api_gateway_number_of_requests_widget(),
            self.create_api_gateway_latency_widget(),
            self.create_api_gateway_errors_widget(),
        )


import aws_cdk
import constructs

from . import rest_api


class RestApiSnsConstruct(rest_api.RestApiConstruct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        error_topic=None,
        rest_api_name=None,
        api=None,
        message=None,
        additional_parameters=None,
        method='POST',
        sns_topic_arn=None,
        **kwargs,
    ):
        super().__init__(
            scope, id,
            error_topic=error_topic,
            api=aws_cdk.aws_apigateway.RestApi(
                scope, 'RestApi',
                rest_api_name=rest_api_name,
                deploy_options=self.get_stage_options(),
            ) if api is None else api,
            **kwargs,
        )
        self.add_method(
            method=method,
            path='SendEvent',
            uri='arn:aws:apigateway:us-east-1:sns:path//',
            success_response_templates={
                "message": 'Message added to SNS topic'
            },
            error_selection_pattern="Error",
            request_parameters={
                'integration.request.header.Content-Type': "'application/x-www-form-urlencoded'"
            },
            request_templates=self.call_sns_publish_api(
                sns_topic_arn=sns_topic_arn,
                additional_parameters=additional_parameters,
                message=message
            ),
        )

    @staticmethod
    def parse_additional_parameters(text):
        if text:
            return f"{text}&"
        return ""

    def call_sns_publish_api(
        self,
        additional_parameters=None,
        sns_topic_arn=None,
        message=None,
    ):
        return (
            "Action=Publish&"
            "Version=2010-03-31&"
            f"Message={message}&"
            f"{self.parse_additional_parameters(additional_parameters)}"
            f"TargetArn=$util.urlEncode('{sns_topic_arn}')"
        )


import aws_cdk
import constructs

from . import well_architected_construct


class DynamodbTableConstruct(well_architected_construct.Construct):

    def __init__(
        self, scope: constructs.Construct, id: str,
            table_name=None,
            error_topic=None,
            partition_key: str=None,
            sort_key: str=None,
            time_to_live_attribute=None,
            stream=None,
            **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )

        self.dynamodb_table = aws_cdk.aws_dynamodb.Table(
            self, id,
            table_name=table_name,
            billing_mode=aws_cdk.aws_dynamodb.BillingMode.PAY_PER_REQUEST,
            partition_key=self.get_dynamodb_key(partition_key),
            sort_key=self.get_dynamodb_key(sort_key),
            stream=stream,
            removal_policy=aws_cdk.RemovalPolicy.DESTROY,
            time_to_live_attribute=time_to_live_attribute,
        )
        self.create_user_errors_alarm()
        self.create_throttles_alarm()

    @staticmethod
    def get_dynamodb_key(key):
        # if key is not None:
        if isinstance(key, str):
            return aws_cdk.aws_dynamodb.Attribute(
                name=key,
                type=aws_cdk.aws_dynamodb.AttributeType.STRING,
            )
        if isinstance(key, aws_cdk.aws_dynamodb.Attribute):
            return key
        return None

    def get_dynamodb_metric(self, metric_name, statistic='sum'):
        return self.dynamodb_table.metric(metric_name=metric_name, statistic=statistic)

    def create_throttles_metric(self, action='Read'):
        return self.get_dynamodb_metric(f'{action}ThrottleEvents')

    def create_throttles_alarm(self):
        return self.create_cloudwatch_alarm(
            id="ReadsWritesThrottled",
            metric=self.cloudwatch_math_sum(
                label="dynamodb_throttles",
                m1=self.create_throttles_metric(),
                m2=self.create_throttles_metric('Write'),
            ),
        )

    def create_user_errors_alarm(self):
        return self.create_cloudwatch_alarm(
            id='UserErrorsGreaterThanZero',
            metric=self.dynamodb_table.metric_user_errors(),
            threshold=0,
        )

    def create_system_errors_alarm(self):
        # creates jsii.errors.JSIIError: Alarms on math expressions cannot contain more than 10 individual metrics
        return self.create_cloudwatch_alarm(
            id='SystemErrorsGreaterThanZero',
            metric=self.dynamodb_table.metric_system_errors_for_operations(),
            threshold=0,
        )

    def create_latency_widget(self):
        return self.create_cloudwatch_widget(
            title="dynamodb_latency",
            left=[
                self.dynamodb_table.metric_successful_request_latency(
                    dimensions_map={
                        'TableName': self.dynamodb_table.table_name,
                        'Operation': action,
                    }
                ) for action in (
                    'GetItem', 'UpdateItem', 'PutItem', 'DeleteItem', 'Query',
                )
            ],
        )

    def create_read_write_capacity_widget(self):
        return self.create_cloudwatch_widget(
            title="dynamodb_consumed_read_write_units",
            stacked=False,
            left=[
                self.get_dynamodb_metric(metric_name, statistic='avg') for metric_name in (
                    'ConsumedReadCapacityUnits', 'ConsumedWriteCapacityUnits',
                )
            ]
        )

    def create_throttles_widget(self):
        return self.create_cloudwatch_widget(
            title="dynamodb_throttles",
            left=[
                self.create_throttles_metric(),
                self.create_throttles_metric('Write')
            ],
        )

    def create_cloudwatch_widgets(self):
        return  (
            self.create_latency_widget(),
            self.create_read_write_capacity_widget(),
            self.create_throttles_widget(),
        )


import constructs
import aws_cdk
import jadecobra.aws.lambda_deployer.deploy_lambda_layer
import os

from . import well_architected_construct


class LambdaFunctionConstruct(well_architected_construct.Construct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        concurrent_executions=None,
        duration=60,
        environment_variables=None,
        error_topic:aws_cdk.aws_sns.Topic=None,
        event_bridge_rule=None,
        function_name=None,
        handler_name='handler',
        lambda_directory='lambda_functions',
        layers:list[str]=None,
        on_success=None,
        on_failure=None,
        retry_attempts=None,
        sns_trigger_topic=None,
        sqs_trigger_queue=None,
        vpc=None,
        **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )
        # handler_name = 'handler' if handler_name is None else handler_name
        function_name = function_name if function_name is not None else id
        self.lambda_function = aws_cdk.aws_lambda.Function(
            self, 'LambdaFunction',
            architecture=aws_cdk.aws_lambda.Architecture.ARM_64,
            code=aws_cdk.aws_lambda.Code.from_asset(f"{lambda_directory}/{function_name}"),
            environment=environment_variables,
            handler=f'{function_name}.{handler_name}',
            layers=self.create_layers(layers),
            on_success=on_success,
            on_failure=on_failure,
            reserved_concurrent_executions=concurrent_executions,
            runtime=aws_cdk.aws_lambda.Runtime.PYTHON_3_9,  # type: ignore
            timeout=aws_cdk.Duration.seconds(duration) if duration else None,
            tracing=aws_cdk.aws_lambda.Tracing.ACTIVE,
            vpc=vpc,
            retry_attempts=retry_attempts,
        )
        self.add_event_bridge_rule(event_bridge_rule)
        self.add_sns_trigger(sns_trigger_topic)
        self.add_sqs_trigger(sqs_trigger_queue)
        self.create_invocations_error_greater_than_2_percent_alarm()
        self.create_invocation_longer_than_1_second_alarm()
        self.create_throttled_invocations_greater_than_2_percent_alarm()

    @staticmethod
    def to_camel_case(text):
        return ''.join(text.title().split('-'))

    def create_layer(self, layer):
        return aws_cdk.aws_lambda.LayerVersion(
            self, f'{self.to_camel_case(layer)}LambdaLayer',
            code=aws_cdk.aws_lambda.Code.from_asset(f"lambda_layers/{layer}"),
            description=f"{layer} Lambda Layer"
        )

    def create_aws_sdk_layer(self):
        layer = 'aws-xray-sdk'
        if not os.path.exists(f'lambda_layers/{layer}'):
            jadecobra.aws.lambda_deployer.deploy_lambda_layer.LambdaLayer(
                dependencies=[layer]
            )
        return [self.create_layer(layer)]

    def create_layers(self, layers):
        result = self.create_aws_sdk_layer()
        try:
            for layer in layers:
                result.append(self.create_layer(layer))
        except TypeError:
            'No additional layers specified'
        return result

    def get_lambda_function_metric(self, metric_name):
        return self.lambda_function.metric(metric_name=metric_name, statistic="sum")

    def create_lambda_error_percentage_metric(self):
        return self.create_cloudwatch_math_expression(
            label="invocations_errored_percentage_last_5_mins",
            expression="(errors / invocations) * 100",
            using_metrics={
                "invocations": self.get_lambda_function_metric('Invocations'),
                "errors": self.get_lambda_function_metric("Errors"),
            },
        )

    def create_lambda_throttled_percentage_metric(self):
        # NOTE: throttled requests are not counted in total number of invocations
        return self.create_cloudwatch_math_expression(
            label="throttled_requests_percentage_last_30_mins",
            expression="(throttles * 100) / (invocations + throttles)",
            using_metrics={
                "invocations": self.get_lambda_function_metric("Invocations"),
                "throttles": self.get_lambda_function_metric("Throttles"),
            },
        )

    def create_invocations_error_greater_than_2_percent_alarm(self):
        return self.create_cloudwatch_alarm(
            id="LambdaInvocationsErrorsGreaterThan2Percent",
            metric=self.create_lambda_error_percentage_metric(),
            threshold=2,
        )

    def create_invocation_longer_than_1_second_alarm(self):
        return self.create_cloudwatch_alarm(
            id="LambdaP99LongDurationGreaterThan1s",
            metric=self.lambda_function.metric_duration(statistic="p99"),
            threshold=1000,
        )

    def create_throttled_invocations_greater_than_2_percent_alarm(self):
        return self.create_cloudwatch_alarm(
            id="LambdaThrottledInvocationsGreaterThan2Percent",
            metric=self.create_lambda_throttled_percentage_metric(),
            threshold=2,
        )

    def create_lambda_error_percentage_widget(self):
        return self.create_cloudwatch_widget(
            title="lambda_error_percentage",
            stacked=False,
            left=[self.create_lambda_error_percentage_metric()],
        )

    def create_lambda_duration_widget(self):
        return self.create_cloudwatch_widget(
            title="lambda_duration",
            left=[
                self.lambda_function.metric_duration(statistic=statistic)
                for statistic in ('p50', 'p90', 'p99')
            ],
        )

    def create_lambda_throttled_percentage_widget(self):
        return self.create_cloudwatch_widget(
            title="lambda_throttle_percentage",
            left=[self.create_lambda_throttled_percentage_metric()],
            stacked=False,
        )

    def create_cloudwatch_widgets(self):
        return (
            self.create_lambda_error_percentage_widget(),
            self.create_lambda_duration_widget(),
            self.create_lambda_throttled_percentage_widget(),
        )

    def add_event_bridge_rule(self, event_bridge_rule):
        try:
            event_bridge_rule.add_target(
                aws_cdk.aws_events_targets.LambdaFunction(
                    self.lambda_function
                )
            )
        except AttributeError:
            return

    def add_sns_trigger(self, sns_trigger_topic):
        try:
            self.lambda_function.add_event_source(
                aws_cdk.aws_lambda_event_sources.SnsEventSource(
                    topic=sns_trigger_topic
                )
            )
        except Exception:
            return

    def add_sqs_trigger(self, sqs_trigger_queue):
        try:
            self.lambda_function.add_event_source(
                aws_cdk.aws_lambda_event_sources.SqsEventSource(
                    queue=sqs_trigger_queue
                )
            )
        except Exception:
            return
        else:
            sqs_trigger_queue.grant_consume_messages(self.lambda_function)

def create_python_lambda_function(
        stack,
        concurrent_executions=None,
        duration=60,
        environment_variables=None,
        error_topic=None,
        event_bridge_rule=None,
        function_name=None,
        lambda_directory=None,
        on_failure=None,
        on_success=None,
        retry_attempts=None,
        sns_trigger_topic=None,
        vpc=None,
    ):
    return LambdaFunctionConstruct(
        stack, function_name,
        concurrent_executions=concurrent_executions,
        duration=duration,
        environment_variables=environment_variables,
        error_topic=error_topic,
        event_bridge_rule=event_bridge_rule,
        function_name=function_name,
        lambda_directory=lambda_directory,
        on_failure=on_failure,
        on_success=on_success,
        retry_attempts=retry_attempts,
        sns_trigger_topic=sns_trigger_topic,
        vpc=vpc,
    )


import aws_cdk
import constructs
import json

from . import api


class RestApiConstruct(api.Api):

    def __init__(
        self, scope: constructs.Construct, id: str,
        error_topic=None,
        rest_api_name=None, api=None,
        **kwargs,
    ):
        super().__init__(
            scope, id,
            error_topic=error_topic,
            api=aws_cdk.aws_apigateway.RestApi(
                scope, 'RestApi',
                rest_api_name=rest_api_name,
                deploy_options=self.get_stage_options(),
            ) if api is None else api,
            **kwargs,
        )

    @staticmethod
    def get_stage_options():
        return aws_cdk.aws_apigateway.StageOptions(
            metrics_enabled=True,
            logging_level=aws_cdk.aws_apigateway.MethodLoggingLevel.INFO,
            data_trace_enabled=True,
            stage_name='prod'
        )

    @staticmethod
    def create_response_parameters(content_type=True, allow_origin=True, allow_credentials=True):
        return {
            'method.response.header.Content-Type': content_type,
            'method.response.header.Access-Control-Allow-Origin': allow_origin,
            'method.response.header.Access-Control-Allow-Credentials': allow_credentials,
        }

    @staticmethod
    def json_content_type():
        return 'application/json'

    def create_json_template(self, template):
        return {self.json_content_type(): template}

    def create_method_response(self, status_code=None, response_model=None):
        return aws_cdk.aws_apigateway.MethodResponse(
            status_code=str(status_code),
            response_parameters=self.create_response_parameters(),
            response_models=self.create_json_template(response_model)
        )

    @staticmethod
    def create_schema(title=None, properties=None):
        return aws_cdk.aws_apigateway.JsonSchema(
            schema=aws_cdk.aws_apigateway.JsonSchemaVersion.DRAFT4,
            title=title,
            type=aws_cdk.aws_apigateway.JsonSchemaType.OBJECT,
            properties=properties
        )

    @staticmethod
    def string_schema_type():
        return aws_cdk.aws_apigateway.JsonSchema(
            type=aws_cdk.aws_apigateway.JsonSchemaType.STRING
        )

    def create_response_model(
        self, rest_api=None, model_name=None, properties=None
    ):
        property_keys = ['message']
        property_keys.append(properties) if properties else None
        return rest_api.add_model(
            model_name,
            content_type=self.json_content_type(),
            model_name=model_name,
            schema=self.create_schema(
                title=model_name,
                properties={key: self.string_schema_type() for key in property_keys},
            )
        )

    def success_response(self, rest_api):
        return (
            200,
            self.create_response_model(
                rest_api=rest_api,
                model_name='pollResponse',
            )
        )

    def error_response(self, rest_api):
        return (
            400,
            self.create_response_model(
                rest_api=rest_api,
                model_name='errorResponse',
                properties='state',
            )
        )

    def create_method_responses(self, rest_api):
        return [
            self.create_method_response(
                status_code=status_code,
                response_model=response_model
            ) for status_code, response_model
            in (
                self.success_response(rest_api),
                self.error_response(rest_api),
            )
        ]

    def add_method(
        self,
        method='POST', path=None,
        uri=None, request_templates=None,
        success_response_templates=None, error_selection_pattern=None,
        request_parameters=None,
    ):
        return self.api.root.add_resource(
            path
        ).add_method(
            method,
            self.create_api_integration(
                uri=uri,
                request_templates=self.create_json_template(request_templates),
                success_response_templates=success_response_templates,
                error_selection_pattern=error_selection_pattern,
                request_parameters=request_parameters,
            ),
            method_responses=self.create_method_responses(self.api)
        )

    def get_integration_options(
        self,
        request_templates=None,
        integration_responses=None,
        request_parameters=None,
    ):
        return aws_cdk.aws_apigateway.IntegrationOptions(
            credentials_role=self.api_gateway_service_role,
            request_parameters=request_parameters,
            request_templates=request_templates,
            passthrough_behavior=aws_cdk.aws_apigateway.PassthroughBehavior.NEVER,
            integration_responses=integration_responses,
        )

    def get_integration_responses(
        self,
        success_response_templates=None,
        error_selection_pattern=None,
    ):
        return [
            self.create_integration_response(
                status_code=200,
                response_templates=success_response_templates,
            ),
            self.create_integration_response(
                status_code=400,
                response_templates={
                    "state": 'error',
                    "message": "$util.escapeJavaScript($input.path('$.errorMessage'))"
                },
                selection_pattern=fr"^\[{error_selection_pattern}\].*",
                separators=(',', ':'),
                response_parameters=self.create_response_parameters(
                    content_type=f"{self.json_content_type()}",
                    allow_origin="'*'",
                    allow_credentials="'true'",
                ),
            )
        ]

    def create_api_integration(
        self, request_templates=None,
        request_parameters=None,
        success_response_templates=None,
        error_selection_pattern=None,
        uri=None,
    ):
        return aws_cdk.aws_apigateway.Integration(
            type=aws_cdk.aws_apigateway.IntegrationType.AWS,
            integration_http_method='POST',
            uri=uri,
            options=self.get_integration_options(
                request_templates=request_templates,
                integration_responses=self.get_integration_responses(
                    success_response_templates=success_response_templates,
                    error_selection_pattern=error_selection_pattern,
                ),
                request_parameters=request_parameters
            ),
        )

    def create_integration_response(
        self, status_code=None, response_templates=None,
        response_parameters=None, selection_pattern=None,
        separators=None
    ):
        return aws_cdk.aws_apigateway.IntegrationResponse(
            status_code=str(status_code),
            selection_pattern=selection_pattern,
            response_templates=self.create_json_template(
                json.dumps(response_templates, separators=separators)
            ),
            response_parameters=response_parameters,
        )


import aws_cdk
import constructs
import aws_cdk.aws_apigatewayv2_alpha
import aws_cdk.aws_apigatewayv2_integrations_alpha

from . import api
from . import lambda_function
from . import well_architected_construct


class ApiLambdaConstruct(well_architected_construct.Construct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        create_http_api=False,
        create_rest_api=False,
        duration=60,
        error_topic=None,
        function_name=None,
        lambda_directory=None,
        event_bridge_rule=None,
        environment_variables=None,
        concurrent_executions=None,
        handler_name='handler',
        layers:list[str]=None,
        on_success=None,
        on_failure=None,
        retry_attempts=None,
        sns_trigger_topic=None,
        sqs_trigger_queue=None,
        vpc=None,
        **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )
        self.lambda_construct = lambda_function.LambdaFunctionConstruct(
            self, 'LambdaFunction',
            concurrent_executions=concurrent_executions,
            duration=duration,
            environment_variables=environment_variables,
            error_topic=error_topic,
            event_bridge_rule=event_bridge_rule,
            function_name=function_name,
            handler_name=handler_name,
            lambda_directory=lambda_directory,
            layers=layers,
            on_success=on_success,
            on_failure=on_failure,
            retry_attempts=retry_attempts,
            sns_trigger_topic=sns_trigger_topic,
            sqs_trigger_queue=sqs_trigger_queue,
            vpc=vpc,
        )
        self.lambda_function = self.lambda_construct.lambda_function
        self.api_construct = self.create_api(
            create_http_api=create_http_api,
            create_rest_api=create_rest_api,
            lambda_function=self.lambda_function,
            error_topic=error_topic,
        )

    def create_api(self,
        create_http_api=None, create_rest_api=None,
        lambda_function=None, error_topic=None
    ):
        if create_http_api:
            return self.create_http_api_lambda(
                lambda_function=lambda_function,
                error_topic=error_topic
            )
        if create_rest_api:
            return self.create_rest_api_lambda(
                lambda_function=lambda_function,
                error_topic=error_topic
            )

    def create_http_api_lambda(
        self, error_topic=None, lambda_function=None
    ):
        return api.Api(
            self, 'HttpApiGateway',
            error_topic=error_topic,
            api_gateway_service_role=False,
            api=aws_cdk.aws_apigatewayv2_alpha.HttpApi(
                self, 'HttpApi',
                default_integration=aws_cdk.aws_apigatewayv2_integrations_alpha.HttpLambdaIntegration(
                    'HttpApiLambdaFunction',
                    handler=lambda_function
                ),
            )
        )

    def create_rest_api_lambda(
        self, error_topic=None, lambda_function=None,
        proxy=True
    ):
        return api.Api(
            self, 'RestApiGateway',
            error_topic=error_topic,
            api_gateway_service_role=False,
            api=aws_cdk.aws_apigateway.LambdaRestApi(
                self, 'RestApiLambdaFunction',
                handler=lambda_function,
                proxy=proxy,
            )
        )


import aws_cdk
import constructs

from . import well_architected_construct


class WebApplicationFirewall(well_architected_construct.Construct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        target_arn=None,
        error_topic=None,
        web_application_firewall_scope='REGIONAL',
        **kwargs
    ) -> None:
        '''NOTE: HTTP APIs are not supported use WAF + CloudFront > HTTP API'''
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )
        self.create_web_application_firewall_association(
            target_arn=target_arn,
            web_application_firewall=self.create_web_application_firewall(
                scope=web_application_firewall_scope,
                web_application_firewall_rules=self.web_application_firewall_rules(),
            ),
        )

    def web_application_firewall_rules(self):
        return [
            self.common_ruleset(1),
            self.anonymous_ip_rule(2),
            self.restricted_ip_list_rule(3),
            self.add_geoblock_rule(4),
        ]

    def create_visibility_configuration(self, metric_name):
        return aws_cdk.aws_wafv2.CfnWebACL.VisibilityConfigProperty(
            cloud_watch_metrics_enabled=True,
            metric_name=metric_name,
            sampled_requests_enabled=True
        )

    @staticmethod
    def create_managed_rule_group_statement(name=None, excluded_rules=None):
        return aws_cdk.aws_wafv2.CfnWebACL.StatementProperty(
            managed_rule_group_statement=aws_cdk.aws_wafv2.CfnWebACL.ManagedRuleGroupStatementProperty(
                name=name,
                vendor_name='AWS',
                excluded_rules=excluded_rules if excluded_rules else []
            )
        )

    def create_managed_rule(self, name=None, priority=None, excluded_rules=None):
        return aws_cdk.aws_wafv2.CfnWebACL.RuleProperty(
            name=name,
            priority=priority,
            override_action=aws_cdk.aws_wafv2.CfnWebACL.OverrideActionProperty(none={}),
            visibility_config=self.create_visibility_configuration(name),
            statement=self.create_managed_rule_group_statement(
                name=name,
                excluded_rules=excluded_rules
            ),
        )

    def common_ruleset(self, priority):
        return self.create_managed_rule(
            name='AWSManagedRulesCommonRuleSet',
            priority=priority,
            excluded_rules=[
                aws_cdk.aws_wafv2.CfnWebACL.ExcludedRuleProperty(
                    name='SizeRestrictions_BODY'
                )
            ],
        )

    def anonymous_ip_rule(self, priority):
        return self.create_managed_rule(
            name='AWSManagedRulesAnonymousIpList',
            priority=priority,
        )

    def restricted_ip_list_rule(self, priority):
        return self.create_managed_rule(
            name='AWSManagedRulesAmazonIpReputationList',
            priority=priority,
        )

    @staticmethod
    def country_codes():
        return ['NZ']

    def add_geoblock_rule(self, priority):
        return aws_cdk.aws_wafv2.CfnWebACL.RuleProperty(
            name='GeoBlockingRule',
            priority=priority,
            action=aws_cdk.aws_wafv2.CfnWebACL.RuleActionProperty(block={}),
            visibility_config=self.create_visibility_configuration('geoblock'),
            statement=aws_cdk.aws_wafv2.CfnWebACL.StatementProperty(
                geo_match_statement=aws_cdk.aws_wafv2.CfnWebACL.GeoMatchStatementProperty(
                    country_codes=self.country_codes(),
                )
            ),
        )

    def create_web_application_firewall(self, web_application_firewall_rules=None, scope=None):
        return aws_cdk.aws_wafv2.CfnWebACL(
            self, 'WebApplicationFirewall',
            default_action=aws_cdk.aws_wafv2.CfnWebACL.DefaultActionProperty(allow={}),
            scope=scope,
            visibility_config=self.create_visibility_configuration('webACL'),
            name='WebApplicationFirewall',
            rules=web_application_firewall_rules
        )

    def create_web_application_firewall_association(self, web_application_firewall=None, target_arn=None):
        return aws_cdk.aws_wafv2.CfnWebACLAssociation(
            self, 'WebApplicationFirewallAPIGatewayAssociation',
            web_acl_arn=web_application_firewall.attr_arn,
            resource_arn=target_arn
        )


from . import api
from . import api_lambda
from . import api_lambda_dynamodb
from . import dynamodb_table
from . import http_api_step_functions
from . import lambda_function
from . import rest_api_sns
from . import rest_api
from . import sns_lambda
from . import web_application_firewall
from . import well_architected_construct


import aws_cdk
import constructs

from . import lambda_function
from . import well_architected_construct


class SnsLambdaConstruct(well_architected_construct.Construct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        sns_topic: aws_cdk.aws_sns.Topic=None,
        error_topic=None,
        function_name=None,
        environment_variables=None,
        lambda_directory=None,
        **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            error_topic=error_topic,
            **kwargs
        )

        self.lambda_function = lambda_function.create_python_lambda_function(
            scope,
            function_name=function_name,
            error_topic=error_topic,
            lambda_directory=lambda_directory,
            environment_variables=environment_variables,
        )
        sns_topic.add_subscription(
            aws_cdk.aws_sns_subscriptions.LambdaSubscription(
                self.lambda_function.lambda_function
            )
        )


import aws_cdk
import aws_cdk.aws_apigatewayv2_alpha
import constructs

from . import api


class HttpApiStepFunctionsConstruct(api.Api):

    def __init__(
        self, scope: constructs.Construct, id: str,
        error_topic=None,
        state_machine_arn=None,
        api_gateway_service_role=None,
        **kwargs,
    ):
        super().__init__(
            scope, id,
            error_topic=error_topic,
            api_gateway_service_role=api_gateway_service_role,
            api=aws_cdk.aws_apigatewayv2_alpha.HttpApi(
                scope, 'HttpApi',
                create_default_stage=True,
            ),
            **kwargs,
        )
        self.create_http_api_step_functions_route(state_machine_arn)

    def create_http_api_step_functions_route(
        self, state_machine_arn
    ):
        return aws_cdk.aws_apigatewayv2.CfnRoute(
            self, 'HttpApiStateMachineDefaultRoute',
            api_id=self.api_id,
            route_key=aws_cdk.aws_apigatewayv2_alpha.HttpRouteKey.DEFAULT.key,
            target=f'integrations/{self.create_http_api_stepfunctions_integration(state_machine_arn)}',
        )

    def create_http_api_stepfunctions_integration(
        self, state_machine_arn
    ):
        return aws_cdk.aws_apigatewayv2.CfnIntegration(
            self, 'HttpApiStateMachineIntegration',
            api_id=self.api_id,
            integration_type='AWS_PROXY',
            connection_type='INTERNET',
            integration_subtype='StepFunctions-StartSyncExecution',
            credentials_arn=self.api_gateway_service_role.role_arn,
            request_parameters={
                "Input": "$request.body",
                "StateMachineArn": state_machine_arn,
            },
            payload_format_version="1.0",
            timeout_in_millis=10000
        ).ref


from time import time
import constructs

from . import api_lambda
from . import dynamodb_table
from . import well_architected_construct


class ApiLambdaDynamodbConstruct(api_lambda.ApiLambdaConstruct):

    def __init__(
        self, scope: constructs.Construct, id: str,
        time_to_live_attribute=None,
        sort_key=None,
        partition_key=None,
        **kwargs
    ) -> None:
        super().__init__(
            scope, id,
            **kwargs
        )
        self.dynamodb_construct = dynamodb_table.DynamodbTableConstruct(
            self, 'DynamoDbTable',
            partition_key=partition_key,
            sort_key=sort_key,
            time_to_live_attribute=time_to_live_attribute,
            error_topic=self.error_topic,
        )
        self.lambda_function.add_environment(
            key='DYNAMODB_TABLE_NAME',
            value=self.dynamodb_construct.dynamodb_table.table_name
        )
        self.dynamodb_construct.dynamodb_table.grant_read_write_data(
            self.lambda_function
        )
